/*
 *    File history:
 *       - created by Jim Battle for decoding Processor Technology PTDOS
 *         disks recorded using the PT Helios disk controller
 *       - Modified substantially to support DG Hard Sector Disks,
 *         and also updated to be compatible with catweasl.c from cw2dmk
 *         version 4.3 (From Tim Mann -- see below).  Also, updated to
 *         support a Catweasel Mk IV.  Jay R. Jaeger, October, 2006
 *       - modified yet again by Jim Battle for use decoding Wang 2200
 *         hard sectored 8" floppies.  As I wanted to fundamentally change
 *         how the disk decoding was done, I removed the parameterization
 *         Jay Jaeger did because I wouldn't have been able to test for
 *         backward compatability, plus there really is little synergy
 *         in making one tool to read these three very different formats.
 *
 *    The PCI scanning and catweasel interface routines are based on:
 *        cw2dmk: Dump floppy disk from Catweasel to .dmk format.
 *        Copyright (C) 2000 Timothy Mann
 *        Id: cw2dmk.c,v 1.28 2003/06/19 05:53:28 mann Exp $
 *
 *    Depends on Linux Catweasel driver code by Michael Krause
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *    USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <assert.h>

#include "cwwvd.h"
#include "cwfloppy.h"
#include "cwpci.h"

// ========================================================================
// catweasel parameters
// ========================================================================

const int side = 0; // doesn't matter for helios

// program operation
const int spinup_ms = 750;  // # of milliseconds to spin up to speed
const int rev_ms    = 167;  // # of milliseconds per revolution

struct catweasel_contr c;   // cached info about controller we're using

// ========================================================================
// dump out raw data samples to a file, along with an index
// ========================================================================

#define FM_DEBUG (get_debug() > 1)  // 1=print diagnostics during fm decode

// ========================================================================
// read real disk and process it a track at a time
// ========================================================================

// how long to capture to the catweasel buffer?
// in the worst case, a track has nothing but 1 bits, so we have:
//    (16 sectors)*(325 bytes/sector)*(8 bits/byte)*(2 transitions/bit)
// or 83200 samples per revolution.  since the catweasel has a 128KB buffer,
// we can store only 1.57 revolutions worth.
//
// on the other hand, the typical case isn't usually that bad, and we'd waste
// time on multiple revolutions waiting to fill the buffer.
//
// we just ask for 1.5 revolutions and then edit the samples to get one clean
// revolution.

#define CAPTURE_TIME ((int)(1.5 * rev_ms))

// read a track and return a complete track, starting at sector 0.
// one complication is that the disk format is hard sectored we can't
// simply ask the catweasel to read index hole to index hole; we must
// use a timed read.

// the sample buffer is passed in, as is an array to hold SECTORS+1 sector
// start points.
// note that the sector_offset[0] most likely isn't zero.
//
// returns 0 on success, -999 if no disk, non-zero on failure.

// read one track off a real disk, then edit the samples such that the first
// sample of logical sector 0 comes first and any extra samples after logical
// sector 16 get dropped.  this may require reordering samples because we have
// no idea what the rotational phase is when we are done stepping to a given
// track.
//
// returns 0 on success, -999 if no disk, non-zero on failure.
// if we are successful, the transition data is saved into *sample in order
// of logical sector 0 to sector 15, and the index to sectors 0 to 16 is
// saved to *sector_offset.

static int
read_track(int drive, int track, uint8 *sample, int *sector_offset)
{
    const int rpm       = 360;        // disk rotational speed
    const int sectors   = HSECTORS;   // hard sectors per track
    const int sector_us = (60 * 1000000) / (rpm * sectors);  // usec/track
    const int sector_ticks = opts.cwmult*sector_us * (CWHZ/1000000.0);  // assuming 2x clock
//printf("sector_ticks=%d, half=%d to %d, full=%d to %d\n",
//        sector_ticks, (int)(0.4*sector_ticks), (int)(0.6*sector_ticks),
//                      (int)(0.8*sector_ticks), (int)(1.2*sector_ticks) );

    // booleans to indicate if the time duration between the start of
    // two index holes corresponds to a full hard sector or a half sector.
    // half sectors appear due to the index hole in the first sector of
    // the track.
    #define HALF_IDX(x) (((x) > sector_ticks*0.4) && ((x) < sector_ticks*0.6))
    #define FULL_IDX(x) (((x) > sector_ticks*0.8) && ((x) < sector_ticks*1.2))
    #define TICKS_TO_US(t) ((int)((t)/(opts.cwmult*CWHZ/1000000.0)))

    // data in the order as it came off the disk
    uint8 RAM[CW_MEMSIZE];

    #define MAX_IDXS 64 // most number of index holes to capture
    struct {
        int idx;           // location of first sample
        int ticks;         // number of ticks until the next index hole
    } idx_list[MAX_IDXS];  // index to where each index hole starts

    struct {
        int idx;         // location of first sample
        int ticks;       // number of ticks until the next index hole
        int samples;     // the number of samples in this sector
    } ls_list[SECTORS];  // index to where each index hole starts

    int no_idx = 0;     // # of times index hole wasn't seen
    int retry;          // # of retries for other reasons

    catweasel_seek(&c.drives[drive], track);

    for(retry = 0; retry < opts.retries; retry++) {

        int first_valid_sample; // first meaningful sample in RAM
        int last_valid_sample;  // last meaningful sample in RAM
        int i, idx_hole_prev, ticks, idx_marks;

        catweasel_fillmem(&c, 0xFF);

        // Read from catweasel, storing index holes in high bit...
        catweasel_read(&c.drives[drive], side, opts.cwmult, CAPTURE_TIME, 1);

        // first, assume the buffer is filled, skipping the first and last
        // few in case of crud
        first_valid_sample = 4;
        last_valid_sample = CW_MEMSIZE-4;

        // figure out how much catweasel captured.
        // catweasel_read() drops a 0x80,0x00 pair to mark the end.
        for(i=0; i<CW_MEMSIZE-8; i++) {
            RAM[i] = catweasel_get_byte(&c);
            if ((i > 1) && (RAM[i-1] == 0x80) && (RAM[i-0] == 0x00)) {
                last_valid_sample = i-2;
                break;
            }
        }

        if (0 && track == 0 && retry == 0) {
            int ii;
            printf("---- RAM dump, track 0, %d valid samples ----\n", last_valid_sample-first_valid_sample+1);
            for(ii=0; ii<= last_valid_sample; ii++) {
                if (ii%16 == 0) { printf("%05x:", ii); }
                printf(" %02x", RAM[ii]);
                if (ii%16 == 15) { printf("\n"); }
            }
        }

        // figure out where the index marks are.  although there are 32
        // physical sectors, the wang format ignores alternate index holes
        // and views the disk as having 16 sectors.  the sector containing
        // the index mark is the first sector of the track.

        ticks = 0;
        idx_marks = 0;
        idx_hole_prev = (RAM[first_valid_sample] & 0x80);

        for(i = first_valid_sample; i <= last_valid_sample; i++) {
            int idx_hole = (RAM[i] & 0x80);
            if (idx_hole && !idx_hole_prev && (idx_marks < MAX_IDXS)) {
                // start of new index hole
                idx_list[idx_marks].idx = i;
                if (idx_marks > 0) {
                    idx_list[idx_marks-1].ticks = ticks;
// printf("found complete sector %d; %d ticks = %d us)\n", idx_marks-1, ticks, TICKS_TO_US(ticks));
//printf("idx_list[%d].ticks = %d\n", idx_marks-1, idx_list[idx_marks-1].ticks);
                }
                idx_marks++;
                ticks = 0;
            }
            ticks += (RAM[i] & 0x7F);
            idx_hole_prev = idx_hole;
        }

        if (get_debug()) {
            // print sector lengths
            printf("  Sector info for track %d\n",track);
            printf("# Sector  time (us)  offset  length\n");
            printf("# ======  =========  ======  ======\n");
            for(i=0; i<idx_marks-1; i++) {
                printf("# %6d  %9d  %6d  %6d\n", i,
                       TICKS_TO_US(idx_list[i].ticks),
                       idx_list[i].idx,
                      (idx_list[i+1].idx - idx_list[i].idx) );
            }
        }

        // sectors nominally come 5.2 ms apart.  the 33rd hole, the index,
        // comes during the middle of the first logical sector.  find these
        // so we can treat them as one logical sector instead of two, and
        // note the first one found.
        {
            int first_sector_idx = -1;
            int error = 0;

            // find where sector 0 begins
            for(i=0; (i < idx_marks-4); i++) {
                if ( HALF_IDX(idx_list[i+0].ticks) &&
                     HALF_IDX(idx_list[i+1].ticks) &&
                     FULL_IDX(idx_list[i+2].ticks) ) {
                    ls_list[0].idx = idx_list[i].idx;
                    ls_list[0].ticks = idx_list[i+0].ticks
                                     + idx_list[i+1].ticks
                                     + idx_list[i+2].ticks;
                    ls_list[0].samples = idx_list[i+3].idx - idx_list[i].idx;
                    first_sector_idx = i;
                    break;
                }
            }
            if (first_sector_idx < 0) {
                no_idx++;
                if (no_idx > 2)
                    return -999;    // no index hole seen
                continue; // recapture track
            }
            // printf("Logical sector 0 begins at hard idx %d\n", first_sector_idx);

            // now scan all the other hard sectors forming the other 15
            // logical sectors.  because we might start the capture at an
            // arbitrary phase, we have to deal with wrapping.
            for(i=1; i<SECTORS; i++) {
                int forward_idx = first_sector_idx+1 + 2*i;
                int reverse_idx = first_sector_idx - 2*(SECTORS-i);
                int idx = (forward_idx+2 < idx_marks) ? forward_idx
                        : (reverse_idx >= 0)          ? reverse_idx
                                                      : -1;
                // printf("Check logical sector %d, begins at hard idx %d\n", i, idx);
                if (idx < 0) {
                    error = 1;
                    if (get_debug())
                        fprintf(stderr, "\ndidn't capture logical sector %d\n", i);
                    printf("\ndidn't capture logical sector %d\n", i);
                    break;
                }
                // make sure we have reasonable looking sector timing
                if (FULL_IDX(idx_list[idx+0].ticks) &&
                    FULL_IDX(idx_list[idx+1].ticks) ) {
                    ls_list[i].idx = idx_list[idx].idx;
                    ls_list[i].ticks = idx_list[idx+0].ticks
                                     + idx_list[idx+1].ticks;
                    ls_list[i].samples = idx_list[idx+2].idx - idx_list[idx].idx;
                } else {
                    // unexpected sector duration
                    error = 1;
                    if (get_debug())
                        fprintf(stderr, "\nsector %d has unexpected duration of (%d+%d) us\n",
                                        i, TICKS_TO_US(idx_list[idx+0].ticks),
                                           TICKS_TO_US(idx_list[idx+1].ticks) );
                    printf("\nsector %d has unexpected duration of (%d+%d) us\n",
                           i, TICKS_TO_US(idx_list[idx+0].ticks),
                              TICKS_TO_US(idx_list[idx+1].ticks) );
                    break;
                }
            } // for(i<SECTORS)

            if (error) {
                continue; // recapture track
            }
        }

        break; // no errors

    } // for(retry < opts.retries)

    // export the samples and index list in logical order, not necessarily
    // the order we captured.
    if (retry < opts.retries) {
        int copy_cnt = 0;
        int i;
        for(i=0; i < SECTORS; i++) {
            // find the location of logical sector i in the original buffer
            // each logical sector is two hard sectors
            // printf("copying logical sector %d, from 0x%x to 0x%x\n", i, ls_list[i].idx, ls_list[i].idx+ls_list[i].samples-1);
            memcpy( &sample[copy_cnt],
                    &RAM[ls_list[i].idx],
                    ls_list[i].samples );
            sector_offset[i] = copy_cnt;
            copy_cnt += ls_list[i].samples;
            if (FM_DEBUG) {
                printf("Sector %d, Offset %d\n", i, sector_offset[i]);
            }
        }
        // this is needed so the caller can determine the length of the last sector
        sector_offset[SECTORS] = copy_cnt;

        if (get_debug()) {
           printf("Debug: Read Track returning success (0)\n");
        }
        return 0;  // success
    }

    if (get_debug()) {
        printf("Debug: Read Track returning failure (1)\n");
    }
    return 1;  // failure
}


static void
cleanup()
{
    catweasel_free_controller(&c);
}

static void
handler(int sig)
{
    cleanup();
    signal(sig, SIG_DFL);
    kill(getpid(), sig);
}


// ========================================================================
// read an actual disk image
// ========================================================================
// attempt to read a real disk via the catweasel board
// returns 0 if OK, 1 if error

int
capture_disk(int drive, svh_disk_t *disk)
{
    int cw_mk = 0;              // Version of catweasel
    int failing_tracks = 0;
    int port = 0;               // only one catweasel in system
    int tick_threshold = 0;     // midpoint between fast and slow transition buckets
    int prev_thold = 0;         // previous working FM decoder threshold

    uint8 RAM[CW_MEMSIZE];      // copy of the catweasel track timing data
    int ch, track;

    /* Keep drive from spinning endlessly on (expected) signals */
    signal(SIGHUP,  handler);
    signal(SIGINT,  handler);
    signal(SIGQUIT, handler);
    signal(SIGPIPE, handler);
    signal(SIGTERM, handler);

    /* Start Catweasel */
    port = pci_find_catweasel(port,&cw_mk);

    if (get_debug()) {
        printf("Catweasel detected at port 0x%x Model %d\n",port,cw_mk);
    }

#if linux
    if ((cw_mk == 1 && ioperm(port,8,1) == -1) ||
        (cw_mk >= 3 && iopl(3) == -1)) {
        fprintf(stderr,"helios: No access to I/O ports\n");
        return 1;
    }
#endif

    ch = catweasel_init_controller(&c, port, cw_mk, getenv("CW4FIRMWARE")) &&
         catweasel_memtest(&c);
    if (ch) {
        if (get_debug()) {
           printf("Detected Catweasel MK%d at port 0x%x\n", cw_mk, port);
           fflush(stdout);
        }
    } else {
        fprintf(stderr, "cw2dmk: Failed to detect Catweasel at port 0x%x\n", port);
        return 1;
    }

    catweasel_detect_drive(&c.drives[drive]);

    /* Error if drive not detected */
    if (c.drives[drive].type == 0) {
        fprintf(stderr, "helios: Failed to detect any drives\n");
        cleanup();
        return 1;
    }

    /* Select drive, start motor, wait for spinup */
    catweasel_select(&c, !drive, drive);
    catweasel_set_motor(&c.drives[drive], 1);
    catweasel_usleep(spinup_ms * 1000);

    fflush(stdout);

    for(track=0; track < opts.tracks; track++) {

        int retry;
        int bad_track = 0;
        int bad_sectors = 0;

        fprintf(stderr, "\rreading track %d", track);

        for(retry=0; (retry <= opts.retries) && !bad_track; retry++) {

            int idx_start[SECTORS+1];
            int sec;
            int failure;

            // read a nicely aligned track
            failure = read_track(drive, track, RAM, idx_start);

            if (failure == -999) {
                fprintf(stderr, "\nNo index hole seen -- is the drive empty?\n");
                exit(-1);   // just quit -- don't return error code
            }

            if (failure) {
                fprintf(stderr, " ... giving up on this track\n");
                bad_track = 1;
                continue;
            }

            // process any sector which hasn't been successfully decoded
            bad_sectors = 0;
            for(sec=0; sec < SECTORS; sec++) {

                if ((disk->trk[track].sector[sec].samples == 0) ||
                    (disk->trk[track].sector[sec].hdr_error) ||
                    (disk->trk[track].sector[sec].data_error)) {

                    int sector_samples = idx_start[sec+1] - idx_start[sec];
                    // copy data to disk data struct for this sector

                    disk->trk[track].sector[sec].samples = sector_samples;
                    if (disk->trk[track].sector[sec].raw != NULL) {
                       free(disk->trk[track].sector[sec].raw);
                    }
                    disk->trk[track].sector[sec].raw = safe_malloc(sector_samples);
                    memcpy( disk->trk[track].sector[sec].raw,
                            &RAM[idx_start[sec]],
                            sector_samples );
if (0 && track == 0) {
    int j;
    printf("decoding track 0, sector %d, raw samples:\n", sec);
    for(j=0; j<sector_samples; j++) {
        printf("%02x ", disk->trk[track].sector[sec].raw[j]);
    }
    printf("\n");
}

                    // decode it
                    sector_decode(disk, track, sec);
                    if (disk->trk[track].sector[sec].hdr_error ||
                        disk->trk[track].sector[sec].data_error) {
                        bad_sectors++;
                    }
if (0 && track == 0) {
    int j;
    printf("decoding track 0, sector %d, data_bytes=%d\n", sec, disk->trk[track].sector[sec].data_bytes);
    for(j=0; j<256; j++) {
        printf("%02x ", disk->trk[track].sector[sec].data[j]);
    }
    printf("\n");
}
                }

            } // for (sec)

            if (bad_sectors == 0)
                break;

        } // for (retry)

        if (bad_track) {
            fprintf(stderr,"Track %d wasn't readable\n", track);
            failing_tracks++;
        } else if (bad_sectors) {
            fprintf(stderr,"Track %d has %d bad sectors\n", track, bad_sectors);
            failing_tracks++;
        }

    } // for (track)

    fprintf(stderr, "\n");
    if (failing_tracks > 0) {
        fprintf(stderr, "Disk read with %d failing tracks\n", failing_tracks);
    }

    cleanup();

    return (failing_tracks != 0);  // OK=0, 1=error
}
