#ifndef _CWWVD_H_
#define _CWWVD_H_

// ========================================================================
// type aliases
// ========================================================================

typedef unsigned char  uint8;
typedef unsigned short uint16;

// ========================================================================
// disk parameters
// ========================================================================

// Disk geometry information

#define HSECTORS  (32)   // number of hard sectors per disk
#define SECTORS   (16)   // number of logical sectors per disk

#define HDR_BYTES  (2)   // number of header bytes per sector
#define DATA_BYTES (256) // number of data bytes per sector
#define CRC_BYTES  (2)   // number of crc bytes per sector

// There are a couple of types defined in here that need static limits
#define MAXSECTORS      (32)
#define MAXTRACKS       (77)

// ========================================================================
// utility functions
// ========================================================================

// just like malloc, but exits with error msg if malloc fails
void* safe_malloc(size_t bytes);

// remove trailing spaces, tabs, carriage returns, and newlines
void strip_line_ending(char *p);

// return NULL on failure, pointer to static buffer otherwise
// maximum length capture is 200 chars.
char *get_first_line_of_file(const char *filename);

// append the source string to the destination string.
// a "\n" is appended to the end of both.
void label_append(char **d, char *s);

// send "LABEL blah blah blah" lines to specified RAW or BYT file
void label_emit(FILE *fp, char *label);

// setter and getter for global debug value
int get_debug();
void set_debug(int value);

// command line options
// this is a global so decoding routines can access tuning parameters
// without having to pass them around
typedef struct {
    int   retries;  // number of attempts to read a given track
    int   tracks;   // number of tracks to dump
    int   skip1;    // number of symbols to skip after sector hole
    int   skip2;    // number of symbols to skip after header decode
    // not exposed -- does it makes sense to?:
    int   cwmult;   // catweasel clock multiplier
                    // sample at 7 MHz * clockmult (1, 2, 4)
} opts_t;

extern opts_t opts;

// ========================================================================
// decoded disk format; can hold .RAW and/or .BYT and .WVD data
// ========================================================================

// store one disk's worth of data
typedef struct {
    struct {
        // raw data:
        int   samples;          // number of raw data samples in this sector
        uint8 *raw;             // catweasel transition data for this sector
        // decoded data:
        int   hdr_error;        // 0=OK, !=0 is error in decoding
        int   hdr_bytes;        // size of header block (0 if none)
        uint8 header[2];        // decoded header bytes
        int   data_error;       // 0=OK, !=0 is error in decoding
        int   data_bytes;       // size of data block (0 if none)
        uint8 data[258];        // decoded body bytes; 256 data, 2 crc
    } sector[MAXSECTORS];
} trk_t;

typedef struct {
    char  *label;       // disk label; can be multiline (\n separators)
    trk_t trk[MAXTRACKS];
} svh_disk_t;

// init and free a decoded disk data structure
void disk_init(svh_disk_t *disk);
void disk_free(svh_disk_t *disk);

// ========================================================================
// various intermodule functions
// ========================================================================

// read real disk and process it a track at a time
// returns 0 if OK, 1 if error
int capture_disk(int drive, svh_disk_t *disk);

// sector level decoding routines
void sector_decode_defaults(void);
int sector_decode(svh_disk_t *disk, int track, int sector);

#endif // _CWWVD_H_
