/*
 *    Modified substantially to support DG Hard Sector Disks,
 *    and also updated to be compatible with catweasl.c from cw2dmk
 *    version 4.3 (From Tim Mann -- see below).  Also, updated to
 *    support a Catweasel Mk IV.  Jay R. Jaeger, October, 2006
 *
 *    This module is responsible for reading the real helios/ptdos
 *    floppy disks via a PCI-attached Catweasel card.  With apologies
 *    to Tim Mann, I didn't carry forward the full generality of his
 *    code since I don't have a 1st generation Catweasel card to test
 *    things, nor am I set up to test things on a linux system.
 *
 *    This module is also responsible for FM decoding the bits on
 *    each track and storing the resulting byte and metadata into
 *    a decoded disk data structure.
 *
 *    This module can also write and read a virtual disk image in
 *    the "raw" format, which is simply the relevant parts of the
 *    the Catweasel sample buffer, track by track.  The reason for
 *    supporting this is that should a disk prove difficult to decode,
 *    it is easier/faster to twiddle the decoding heuristics from the
 *    virtual disk image instead of having to sample the physical disk
 *    for each run of the program.
 *
 *    The PCI scanning and catweasel interface routines are based on:
 *        cw2dmk: Dump floppy disk from Catweasel to .dmk format.
 *        Copyright (C) 2000 Timothy Mann
 *        Id: cw2dmk.c,v 1.28 2003/06/19 05:53:28 mann Exp $
 *
 *    Depends on Linux Catweasel driver code by Michael Krause
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *    USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <math.h>   // for pow()

#include "cwwvd.h"
#include "cwfloppy.h" // for CWHZ

// TODO:
//    -- better error reporting
//    -- better progress reporting
//    -- heuristics to tune skip1/skip2 if required

// ========================================================================
// FM decoding logic
// decode and save logical data.
//
// this code is very heuristic; I'm sure it will fail for some disks and
// will require hand tuning.  There are a lot of possible improvements,
// but until I find a disk that requires it, I haven't bothered.
// ========================================================================

#define FM_DEBUG (get_debug() > 1)      // 1=print diagnostics during fm decode

// The Wang 2270 8" disk controller uses disks with 32 hard sectors,
// but it ignores every other index hole, leaving it with 16 sectors
// per track.
//
// At the start of each detected index hole (if it is not an ignored one),
// the controller writes, in order
//      20  '00' bytes
//       1  '03' sync byte
//       1  track byte
//       1  sector byte
//      20  '00' bytes
//     256  data bytes
//       2  crc bytes

// decoding phases
enum { FM_SKIP1=0,      // immediately after sector start
       FM_PREAMBLE1,    // scanning zeros before
       FM_SYNC1,        // parsing the pre-header sync byte
       FM_HEADER,       // reading header data
       FM_SKIP2,        // time killing to skip over crossover noise
       FM_PREAMBLE2,    // scanning zeros before data
       FM_SYNC2,        // parsing the pre-data sync byte
       FM_DATA,         // parsing the data or CRC portion
       FM_POSTAMBLE     // trailing junk
     } fm_state;

static char *fm_name[] = {
       "FM_SKIP1",
       "FM_PREAMBLE1",
       "FM_SYNC1",
       "FM_HEADER",
       "FM_SKIP2",
       "FM_PREAMBLE2",
       "FM_SYNC2",
       "FM_DATA",
       "FM_POSTAMBLE"
    };

// set the tuning parameter defaults
void
sector_decode_defaults(void)
{
    opts.skip1 =  76;   // in terms of bit times
    opts.skip2 = 140;   // in terms of bit times
    opts.cwmult =  2;   // sample data at CWHZ*cwmult
}

// ========================================================================
// histogram analysis, modified from Tim Mann's testhist.c program:
//    From cwfloppy-0.2.1 package, copyright (c) 1998 Michael Krause
//    Modified by Timothy Mann
//    Id: testhist.c,v 1.11 2002/12/08 05:36:47 mann Exp $
// ========================================================================

// analyze the histogram of the samples taken and return the two main peaks,
// which correspond to the fast clock transition and the slow clock transition.
// we return 0 if everything is OK, and 1 if there seems to be a problem.
// problems could be:
//     there aren't exactly two peaks
//     the peaks are not well defined
//     the slow pulse isn't roughly 2x the fast pulse
//     the slow pulse isn't roughly 4 us
static int
histo(int *fast, int *slow, uint8 *sample, int num_samples)
{
    unsigned int hist[128];
    double peak[3], sd[3], ps[3];
    int i, ii, pk;

    for(i=0; i<3; i++) {
        peak[i] = sd[i] = ps[i] = 0.0;
    }

    memset(hist, 0, sizeof(hist));
    for(i=0; i<num_samples; i++) {
        hist[(int)(sample[i]&0x7F)]++;
    }

    /* Print histogram */
    if (0) {
        printf("\n");
        for(i=0;i<128;i+=8) {
        printf("%3d: %06d %06d %06d %06d %06d %06d %06d %06d\n", i,
               hist[i+0], hist[i+1], hist[i+2], hist[i+3],
               hist[i+4], hist[i+5], hist[i+6], hist[i+7]);
        }
    }

    #define bucket(n) (((n) < 0) ? 0 : ((n) > 127) ? 0 : hist[n])
    // do really simple lowpass filter
    if (1) {
        int h[128];
        for(i=0; i<128; i++) {
            h[i] = 1*bucket(i-2) +
                   3*bucket(i-1) +
                   9*bucket(i  ) +
                   3*bucket(i+1) +
                   1*bucket(i+2) ;
        }
        for(i=0; i<128; i++) {
            hist[i] = (int)((1.0/17.0) * h[i] + 0.5);
        }
    }

    /* Print histogram */
    if (0) {
        printf("\n");
        for(i=0;i<128;i+=8) {
        printf("%3d: %06d %06d %06d %06d %06d %06d %06d %06d\n", i,
               hist[i+0], hist[i+1], hist[i+2], hist[i+3],
               hist[i+4], hist[i+5], hist[i+6], hist[i+7]);
        }
    }

    /* find the tallest peak and wipe out any bucket with under 5% of max */
    if (1) {
        int max = 0;
        for(i=0; i<128; i++) {
            if (hist[i] > max)
                max = hist[i];
        }

        for (i=0; i<128; i++) {
            if (20*hist[i] < max)
                hist[i] = 0;
        }
    }

    /* Print histogram */
    if (0) {
        printf("\n");
        for(i=0;i<128;i+=8) {
        printf("%3d: %06d %06d %06d %06d %06d %06d %06d %06d\n", i,
               hist[i+0], hist[i+1], hist[i+2], hist[i+3],
               hist[i+4], hist[i+5], hist[i+6], hist[i+7]);
        }
    }

    /* Find two or three peaks */
    i = 0;
    for (pk=0; pk<3; ) {
        int pwidth = 0;
        int psamps = 0;
        int psampsw = 0;
        while (hist[i] < 64 && i < 128) i++;
        while (hist[i] >= 64 && i < 128) {
            pwidth++;
            psamps  += hist[i];
            psampsw += hist[i] * i;
            i++;
        }
        if (pwidth == 0 || pwidth > 16) {
            /* Not a real peak */
            break;
        }
        peak[pk] = ((double) psampsw) / psamps;
        ps[pk] = psamps;
        sd[pk] = 0.0;
        for (ii = i - pwidth; ii < i; ii++) {
            sd[pk] += hist[ii] * pow((double)ii - peak[pk], 2);
        }
        sd[pk] = sqrt(sd[pk]/((double)psamps-1));
        // fprintf(stderr, "peak %d: mean %.05f, sd %.05f\n", pk, peak[pk], sd[pk]);
        pk++;
    }

    // i've seen disks where there was a broad sweep, maybe due to
    // different rotational speeds when the header was written from
    // when the data was written.  this triggered more than two peaks.
    // here, we try to handle that case by picking the two strongest.
    if (pk != 2) {
#if 0
        fprintf(stderr,"histo: Error: pk is %d\n",pk);
        for(i=0;i<3;i++)
            fprintf(stderr, "%d: ps[]=%f, peak[]=%5.1f\n", i, ps[i], peak[i]);
        for(i=0;i<128;i+=8) {
            fprintf(stderr, "%3d: %06d %06d %06d %06d %06d %06d %06d %06d\n", i,
                            hist[i+0], hist[i+1], hist[i+2], hist[i+3],
                            hist[i+4], hist[i+5], hist[i+6], hist[i+7]);
        }
        return 1;   // treat it as an error; FM should produce two clear peaks
#else
        if ( (ps[0] < ps[1]) && (ps[0] < ps[2]) ) {
            // peak 0 loses
            peak[0] = peak[1]; peak[1] = peak[2];
              sd[0] =   sd[1];   sd[1] =   sd[2];
        } else if ( (ps[1] < ps[0]) && (ps[1] < ps[2]) ) {
            // peak 1 loses
            peak[1] = peak[2];
              sd[1] =   sd[2];
        } // else peak 2 loses
#endif
    }

    // report the peaks, with 0 for non-existant peaks
    *slow = *fast = 0;
    if (pk > 0)
        *fast = (int)floor(peak[0] + 0.5);
    if (pk > 1)
        *slow = (int)floor(peak[1] + 0.5);

    // now sanity check what we got back

    // for clockmult=2, typically sd[0] < 1.8, sd[1] < 2.5
    if (sd[0] > 2.5 || sd[1] > 3.5) {
        fprintf(stderr,"\nhisto_track: sd value peaks not well defined, "
                       "sd[0]: %5.1f, sd[1]: %5.1f\n",sd[0],sd[1]);
        return 1;   // peaks aren't well defined
    }

    if ((2*peak[0] < 0.9*peak[1]) ||
        (2*peak[0] > 1.1*peak[1])) {
        fprintf(stderr,"\nhisto_track: ratio of peaks not close to 2x, "
                       "peak[0]: %5.1f, peak[1]: %5.1f\n",peak[0],peak[1]);
        return 1;   // slow should be close to 2x fast
    }

    return 0;
}


// ========================================================================
// foo
// ========================================================================

// decode bits
//
// The disk data is FM encoded, where there are two clocks (start and middle)
// during a "1" bit and just one clock (start) during a "0" bit.
// Bits are encoded at 250 KHz nominal, which is 4 uS, which comes out to
// (4*CWHZ*clockmult) ticks per bit.  With a clockmult of two, that is 56
// ticks per bit.
//
// we return 0 on success, and non-zero on failure.

// module repair guide #2 gives ucode listing for PCS-II disk controller.
// the write routine first reads the track & sector header bytes, compares
// them (ucode addr 058-074).  at 1.6 uS per instruction, this takes about 
// 20*1.6uS, or 32 uS.  the routine at 17B turns off the read gate, sets
// the write data register to 0, and then kills 460 uS (listed as MSEC)
// before turning on the write gate.  this means the crossover will come
// around 32 + 3*1.6 + 0x90*2*1.6, or about 500 uS.  At an FM rate of
// 250 Kbps (4uS/bit), that is 125 bits.  It seems to then go to 1D3-1D7,
// where it kills another 460 uS (another 115 bit times), before loading
// the RAM address pointer to 1F9 and jumping back to 181.  This is because
// the data rate for 5.25" floppies is half speed (125 Kbps).
// 1F9-1FE is six 00 bytes, then an 03 byte, then the data, so it seems
// like this is how many zero bits we could count on being reasonable
// before the data starts.
//
// nominally the bit times are 27 counts and 55 counts.
// sector 5, I see the track 0, sector 5 bytes fine, then
//    128 '0' bits
//    43, 18, 15, 27 durations
//        43 is really in the middle between short and long,
//        18 and 15 are extremely short
//     55 '0' bits
//     '03' byte
//
// the real hardware does reads like this:
//     do all the preliminary crud to start reading sectors
//     once the right track/sector numbers are found, then
//     wait 0xB0 3.2 uS loops, or 563 uS, which is 140 bit times.
//     then reset CRC and wait for sync byte.
// further, when reading the header, the PCS-II code waits for
// 0xC0*3.2 uS after the end of the sector mark before turning on
// the read gate, or 76 bit times.

static int
decode_sector_data(svh_disk_t *disk, int track, int sector, int tick_threshold)
{
    // create some aliases to reduce code clutter
    int num_samples = disk->trk[track].sector[sector].samples;
    uint8 *sample = disk->trk[track].sector[sector].raw;
    uint8 *header = disk->trk[track].sector[sector].header;
    uint8 *data   = disk->trk[track].sector[sector].data;

    int fm_data;             // accumulated bits of a byte
    int fm_bits;             // count bits received
    int fm_bytes;            // count bytes received
    int fm_preamble_zeros;   // count how many sync zeros we got
    int fm_error;            // unrecoverable error
    int fm_prev_short;       // used for grouping edges in "1" bits

    int any_err = 0;
    int index_bits_seen = 0;
    int mapped_sector = -1;
    int i;

//  if (FM_DEBUG)
//      printf("# ====== track %d, sector %d contains %d samples ======\n",
//              track, sector, num_samples);

    if (FM_DEBUG)
        printf("# starting decode of track %d, sector %d\n", track, sector);

    disk->trk[track].sector[sector].hdr_error = 1;  // until proven otherwise
    disk->trk[track].sector[sector].hdr_bytes = 0;
    disk->trk[track].sector[sector].data_error = 1;
    disk->trk[track].sector[sector].data_bytes = 0;

    fm_error = 0;
    fm_state = FM_SKIP1;
    fm_bytes = fm_bits = fm_preamble_zeros = 0;
    fm_prev_short = 0;

    for(i=0; i<num_samples; i++) {

        int samp = (sample[i] & 0x7F);
        int big = (samp > tick_threshold);

        if (FM_DEBUG) {
            printf("%s: samp=%d, %s\n", fm_name[fm_state], samp,
                    big ? "long" : "short");
        }

        // if we already gave up, don't try to sync again
        if (fm_error)
            return;

        switch (fm_state) {

            // we skip the first 76 bit times after the start of sector to
            // avoid glitches from when the write head was energized when
            // the sector was first written; we skip the first 140 bit times
            // after the header data is decoded to avoid the crossover glitch
            // from when the data preamble bits switch from those written during
            // formatting and those written when the sector data was written.
            // empirically, the preamble1 glitch is about 50 bit times in,
            // although it can be later, and the crossover glitch occurs about
            // 128 bit times after the header data ends.
            case FM_SKIP1:
            case FM_SKIP2:
                // give credit for any contiguous sync 0s that occur at the
                // end of the skip phase leading into the preamble sync phase
                if (big) {
                    fm_preamble_zeros++;
                } else {
                    fm_preamble_zeros = 0;
                }

                // count how much time has passed; we count half bit times
                fm_bits += (big) ? 2 : 1;
                if ((fm_state == FM_SKIP1) && (fm_bits >= opts.skip1*2)) {
                    fm_bytes = 0;
                    fm_bits = fm_preamble_zeros;
                    fm_state = FM_PREAMBLE1;
                } else if ((fm_state == FM_SKIP2) && (fm_bits >= opts.skip2*2)) {
                    fm_bytes = 0;
                    fm_bits = fm_preamble_zeros;
                    fm_state = FM_PREAMBLE2;
                }
                break;

            case FM_PREAMBLE1:
            case FM_PREAMBLE2:
                if (big) {
                    // count how many zero bits we get in a row
                    fm_bits++;
                } else if (fm_bits < 16) {
                    if (FM_DEBUG)
                        printf("\n# lost %s sync after %d bits\n",
                                fm_name[fm_state], fm_bits);
                    fm_bits = 0; // lost sync
                } else {
                    // we just got a short pulse after a train of long pulses
                    if (FM_DEBUG)
                        printf("# leaving %s sync after %d bits\n",
                                fm_name[fm_state], fm_bits);
                    fm_state = (fm_state == FM_PREAMBLE1) ? FM_SYNC1 : FM_SYNC2;
                    fm_bits = fm_bytes = 0;
                    fm_prev_short = 1;  // we just saw it
                }
                break;

            // we enter this state after already seeing the first short pulse
            // the sync pattern (two '1' bits).
            case FM_SYNC1:
            case FM_SYNC2:
                if (big == 0) {
                    fm_prev_short++;
                    if (fm_prev_short == 4) {
                        // we just saw the '03' bit pattern
                        fm_prev_short = 0;
                        fm_state = (fm_state == FM_SYNC1) ? FM_HEADER : FM_DATA;
                        if (FM_DEBUG)
                            printf("# transition to %s\n", fm_name[fm_state]);
                    } // else stay in this state
                } else {
                    // long pulse
                    if (FM_DEBUG)
                        printf("# lost %s sync after %d short bits\n",
                                fm_name[fm_state], fm_prev_short);
                    fm_prev_short = 0;
                    fm_bits = 1;
                    if (fm_state == FM_SYNC1) {
                        // this seems like it shouldn't happen since it is
                        // written in one go at the time of formatting, but
                        // it does appear to sometimes help on actual disks
                        fm_state = FM_PREAMBLE1;
                    } else {
                        // there is often a discontinuity in the PREAMBLE2 window
                        // because the region was initially filled with zeros
                        // on format, but subsequent writes will start filling
                        // at an indeterminate point
                        fm_state = FM_PREAMBLE2;
                    }
                }
                break;

            // we are gathering bytes.  the header contains just two bytes,
            // and the data portion should always have 256 bytes.
            case FM_HEADER:
            case FM_DATA:
                if (big && fm_prev_short) {
                    // oops -- got a runt pulse
                    if (FM_DEBUG)
                        printf("# ERROR: illegal transition (unpaired short) after "
                               "%d bytes, %d bits\n", fm_bytes, fm_bits);
                    fm_error = 1;
                } else if (!big) {
                    if (!fm_prev_short) {
                        // this is the first half of a '1' bit, presumably
                        fm_prev_short = 1;
                        break;
                    }
                    fm_prev_short = 0;  // it has been matched now
                }

                // we finished a 0 bit or a 1 bit
                fm_data = (fm_data << 1) | (!big);  // add data bit
                fm_bits++;

                if (fm_bits < 8)
                    break;  // keep accumulating

                fm_data &= 0xFF;
                if (FM_DEBUG) {
                    if (fm_data >= 32 && fm_data < 128)
                        printf("%d: %02X ('%c')\n", fm_bytes, fm_data, (char)fm_data);
                    else
                        printf("%d: %02X\n", fm_bytes, fm_data);
                    if (fm_bytes%16 == 15)
                        printf("\n");
                }
                data[fm_bytes] = fm_data;
                fm_bytes++;
                fm_data = 0x00;
                fm_bits = 0;

                if ((fm_state == FM_HEADER) && (fm_bytes == 2)) {
                    // printf("\n# this: sector %d, track %d\n", data[1], data[0]);
                    disk->trk[track].sector[sector].hdr_error = fm_error;
                    disk->trk[track].sector[sector].hdr_bytes = 2;
                    disk->trk[track].sector[sector].header[0] = data[0];
                    disk->trk[track].sector[sector].header[1] = data[1];
                    fm_state = FM_SKIP2;
                    fm_bits  = 0;
                    fm_bytes = 0;
                    fm_preamble_zeros = 0;
                    break;
                }

                if ((fm_state == FM_DATA) && (fm_bytes == 258)) {
                    // 256 data bytes plus two CRC bytes

                    // the 2270A controller uses the 74F401 for the CRC.
                    // it implements either the CRC-16 or the CCITT-16.
                    //    ctrl = 000: CRC-16,    x^16 + x^15 + x^2 + 1  (crc reg 0000 at start)
                    //    ctrl = 110: CRC-CCITT, x^16 + x^12 + x^5 + 1  (crc reg FFFF at start)
                    unsigned short crc, crc2;
                    const char *syncbyte = "\03";
                    // after a train of '00' bytes, an '03' byte is used to
                    // synchronize the data stream.  this byte is part of the crc.
                    crc  = calc_crc_16_ibm(0x0000, syncbyte, 1);
                    // CRC the 256 byte data block
                    crc  = calc_crc_16_ibm(crc, &data[0], 256);
                    // continue on with the stored CRC
                    crc2 = calc_crc_16_ibm(crc, &data[256], 2);
                    if (crc2 != 0x0000) {
                        fm_error = 1;
//fprintf(stderr, "# sector %d, crc = %04X, expected %02X%02X\n",
//        sector, crc, data[256], data[257]);
                        if (FM_DEBUG) {
                            printf("# crc = %04X, expected %02X%02X\n",
                                    crc, data[256], data[257]);
                            fprintf(stderr, "# Error: crc = %04X, "
                                    "sector %d, state=%s\n",
                                    crc2, sector, fm_name[fm_state]);
                        }
                    }
                    disk->trk[track].sector[sector].data_error = fm_error;
                    disk->trk[track].sector[sector].data_bytes = 256;

                    fm_state = FM_POSTAMBLE;
                    fm_bytes = 0;
                }
                break;

            case FM_POSTAMBLE:
                return;     // ignore data until the next sector mark

            default:
                assert(0);
        } // switch (bit serial decode)

        any_err |= fm_error;

    } // for (num_samples)

    if (FM_DEBUG)
        printf("\n"); // blank line after sector's data

    return any_err;
}


// ========================================================================
// sector decode boss
// ========================================================================
// attempt to decode one sector's worth of data from the raw samples provided.
// the actual decode is done elsewhere; this is just the tuning supervisor.
// the raw samples come from the *disk structure, and the results are stuffed
// back into the *disk structure.
//
// we may need to tune various parameters to get a clean decode.
// a few different schemes of increasing desperation and cleverness are
// attempted before giving up.
//
//   (1) attempt to decode using the threshold one would expect for the disk
//   (2) try sweeping the threshold -10% to +10%, 2% at a time
//   (3) (NOT IMPLENTED YET)
//       there is a "crossover" during the second preamble where the
//       bits written during formatting get overwritten by the bits
//       written during the data write.  This can sometimes lead to
//       a false sync.  I'm not sure how the real hardware avoids
//       getting confused by it, but the approach here is that if
//       the CRC is bad, go back and try again ignoring the putative
//       false sync before the data bits.
//   (4) (NOT IMPLEMENTED) attempt some bit length compensation to
//       account for the effect of the sequence of pulse lengths
//
// return 0 on success, 1 on failure.

int
sector_decode( svh_disk_t *disk,   // source of raw, dest of decoded data
               int track,          // track number we're decoding
               int sector )        // sector number we're decoding
{
    // calculate where peaks are expected
    //    360 RPM = 6 RPS = 96 sectors/sec
    //    each sector has 20+1+2+20+1+256+2 bytes, or 302 bytes
    //    this means about 2416 bits/sector
    //    250 Kbps; 0 = 4 us, 1 = 2us+2us
    int fast = (int)(CWHZ * opts.cwmult * 2 / 1E6 + 0.5);
    int slow = (int)(CWHZ * opts.cwmult * 4 / 1E6 + 0.5);
    int center_thold;
    int delta, dir;
    int variance;

    // determine bit timing from histogram of data
    if (0) {
        int hfast, hslow;
        int failure = histo(&hfast, &hslow,
                            disk->trk[track].sector[sector].raw,
                            disk->trk[track].sector[sector].samples);
        if (!failure) {
            int thold = (hfast+hslow+1)/2;
            printf("histogram predicts thold of %d\n", thold);
            fast = hfast;
            slow = hslow;
        }
    }

    // sweep the threshold near where we expect it
    center_thold = (fast+slow+1)/2;
    variance = (int)(center_thold / 10.0 + 0.5);
    variance = 0;  // this doesn't seem to help

    for(delta = 0; delta <= variance; delta++) {
        for(dir = -1; dir <= 1; dir += 2) {
            int thold = center_thold + dir*delta;
            int failure = decode_sector_data(disk, track, sector, thold);
            if (!failure) {
                if (delta > 0) fprintf(stderr, "\nsweep delta=%d\n", delta);
                return 0;  // success
            }

            // no need to do both +/- 0
            if (variance == 0)
                break;
        }
    }

    return 1;  // failure
}
