/* read and write .raw disk image to disk
 * this contains the label metadata and all the track transition data
 *
 *   .raw files have this structure
 *   # blah blah blah             (comments)
 *   HELIOS raw                   (first line)
 *   LABEL first line of label    (optional)
 *   LABEL second line of label   (optional)
 *   TRACK t n                    (t=track number)
 *   INDEX s u o n                (s=sector, u=duration in us, o=offset, n=samples)
 *   hh hh hh hh ...              (n raw samples data in hex)
 */

/* read and write ".BYT" format files.
   BYT files are a helios disk image, containing all the decoded bytes
   for each header and data block, along with a bit of metadata.

   BYT is useful while working on routines for converting to other formats
   or debugging the PTDOS (and maybe CP/M in the future) file extraction
   code.  FM decoding the large RAW files takes a fair amount of time on
   my poor old CPU.

   BYT files have this structure:
     HELIOS bytes                 (first line)
     # blah blah blah             (optional comments)
     LABEL first line of label    (optional)
     LABEL second line of label   (optional)
     HEADER t s b e               (t=track, s=sector, b=bytes, e=error)
     hh hh hh hh ...              (n hex bytes of header)
     DATA t s b e                 (t=track, s=sector, b=bytes, e=error)
     hh hh hh hh ...              (n hex bytes of data)
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "cwwvd.h"
#include "fileops.h"
#include "cwfloppy.h"   // needed for CWHZ and CW_MEMSIZE

// ========================================================================
// dump out raw data samples to a file, along with an index
// ========================================================================

int
write_raw_file(const char *filename, svh_disk_t *disk)
{
    int track;
    FILE *fp;

    assert(filename != NULL);

    fp = fopen(filename, "w");
    if (fp == NULL) {
        fprintf(stderr, "Error opening file '%s' for writing\n", filename);
        return 1;
    }
    fprintf(fp, "HELIOS raw\n");
    label_emit(fp, disk->label);
    fprintf(fp, "# clockmult = %d\n", opts.cwmult);

    for(track=0; track<opts.tracks; track++) {

        // how many samples are there on this track?
        int sec;
        int tot_samples = 0;
        for(sec=0; sec<SECTORS; sec++)
            tot_samples += disk->trk[track].sector[sec].samples;
        fprintf(fp, "TRACK %d %d\n", track, tot_samples);

        // print sector lengths
        fprintf(fp, "# Sector  time (us)  offset  length\n");
        fprintf(fp, "# ======  =========  ======  ======\n");
        tot_samples = 0;
        for(sec=0; sec<SECTORS; sec++) {
            int ticks = 0;
            int offset = tot_samples;
            int length = disk->trk[track].sector[sec].samples;
            int j;
            tot_samples += length;
            for(j=0; j<length; j++)
                ticks += (disk->trk[track].sector[sec].raw[j] & 0x7F);
            fprintf(fp, "INDEX %2d  %9d  %6d  %6d\n", sec,
                        (int)(ticks/(opts.cwmult*CWHZ/1000000.0)),
                        offset, length);
        }
        fprintf(fp, "\n");

        // dump samples
        tot_samples = 0;
        for(sec=0; sec<SECTORS; sec++) {
            int offset = tot_samples;
            int length = disk->trk[track].sector[sec].samples;
            int j;
            for(j=0; j<length; j++) {
                fprintf(fp, "%02X ", disk->trk[track].sector[sec].raw[j]);
                if (((tot_samples + j)&15) == 15)
                    fprintf(fp, "\n");
            }
            tot_samples += length;
        }

        fprintf(fp,"\n\n");

    } // for(track)

    fclose(fp);
    return 0;
}

// ========================================================================
// read .raw disk image from file, and decode as we go
// ========================================================================

// read the first line of the file and attempt to classify it
int
file_is_raw(const char *filename)
{
    char *line = get_first_line_of_file(filename);
    return (line != NULL) && (strcmp(line,"HELIOS raw") == 0);
}

int
read_raw_file(const char *filename, svh_disk_t *disk)
{
    int bad_tracks = 0;
    int track;
    FILE *fp;

    if (!file_is_raw(filename))
        return 1;

    fp = fopen(filename, "r");
    assert(fp != NULL);

    opts.tracks = 0;
    for(track=0; track<MAXTRACKS; track++, opts.tracks++) {

        uint8 RAM[CW_MEMSIZE];
        int  idx_start[SECTORS+1];  // starting offset of given track
        char line[300];             // holds line to parse
        int  failure = 0;           // something has gone wrong
        int  good_thold = 0;        // best short/long threshold
        int  bad_sectors = 0;
        int  sec;
        int  i;

        // scan for start of specified track
        while (1) {
            int trk;
            if (fgets(line, sizeof(line)-1, fp) == NULL) {
                // hit the end of the file -- we are done
                fclose(fp);
                return 0;
            }
            if ( (sscanf(line, "TRACK %d", &trk) == 1) &&
                 (trk == track) )
                break; // found the start of the track
            if (strncmp(line, "LABEL ", 6) == 0) {
                strip_line_ending(line);
                label_append( &(disk->label), &line[6] );
                continue;
            }
        }
        fprintf(stderr, "\rreading track %d", track);

        // scan for the sector index table
        for(i=0; i<SECTORS; i++) {
            int p1, p2, p3;  // sector, offset, length
            while (1) {
                if (fgets(line, sizeof(line)-1, fp) == NULL) {
                    fprintf(stderr, "\nError: can't read index table entry %d for track %d\n", i, track);
                    fclose(fp);
                    return 1;
                }
                if (sscanf(line, "INDEX %d %*d %d %d", &p1, &p2, &p3) == 3)
                    break;
            }
            if (p1 != i) {
                fprintf(stderr, "\nError: index table in wrong order for track %d\n", track);
                fclose(fp);
                return 1;
            }
            idx_start[i  ] = p2;
            idx_start[i+1] = p2+p3;
        }

        // pull down sample bytes
        for(i=0; i<idx_start[SECTORS]; i++) {
            int byte;
            char p[300];
            if (fscanf(fp, "%s", p) != 1) {
                fprintf(stderr, "\nError: couldn't read byte %d of track %d\n", i, track);
                fclose(fp);
                return 1;
            }
            byte = (int)strtol(p, (char**)NULL, 16);
            if (byte < 0 || byte > 255) {
                fprintf(stderr, "\nError: byte %d of track %d is out of range\n", i, track);
                fclose(fp);
                return 1;
            }
            RAM[i] = (uint8)byte;
        }

        // copy the raw sector data the disk struct
        for(sec=0; sec<SECTORS; sec++) {
            int samples = idx_start[sec+1] - idx_start[sec];
            assert(idx_start[sec] >= 0);
            assert(idx_start[sec] < idx_start[sec+1]);
            assert(idx_start[sec+1] < CW_MEMSIZE);
            disk->trk[track].sector[sec].samples = samples;
            disk->trk[track].sector[sec].raw = safe_malloc(samples);
            memcpy(disk->trk[track].sector[sec].raw, &RAM[idx_start[sec]], samples);
        }

        // decode raw data
        for(sec=0; sec<SECTORS; sec++) {
            sector_decode(disk, track, sec);
            if (disk->trk[track].sector[sec].hdr_error ||
                disk->trk[track].sector[sec].data_error)
                    bad_sectors++;
        }
        if (bad_sectors) {
            fprintf(stderr,"\rTrack %d has %d bad sector%s\n",
                    track, bad_sectors, (bad_sectors>1)?"s":"");
            bad_tracks++;
        }

    } // for (track)

    fprintf(stderr, "\r");
    if (bad_tracks > 0) {
        fprintf(stderr, "Read %d tracks, with %d failing tracks\n",
                opts.tracks, bad_tracks);
    } else {
        fprintf(stderr, "Read %d tracks successfully\n", opts.tracks);
    }

    return (bad_tracks != 0);  // OK=0, 1=error
}


// ========================================================================
// read .byt disk image from file and process it
// return 0 on success, !=0 on failure
// ========================================================================

// read the first line of the file and attempt to classify it
int
file_is_byt(const char *filename)
{
    char *line = get_first_line_of_file(filename);
    return (line != NULL) && (strcmp(line,"HELIOS bytes") == 0);
}

// read either a HEADER or DATA block, returning a malloc'd
// array with the contents along with meta data.
// return 0 if OK.
static int
read_byt_block(FILE *fp, char *line, svh_disk_t *disk)
{
    char tag[20];       // HEADER or DATA
    char p[300];        // buffer for holding scanf
    char mem[256];      // header or data block
    int t, s, b, e;     // parsed parameters
    int i;

    if (sscanf(line, "%s %d %d %d %d", tag, &t, &s, &b, &e) != 5) {
        fprintf(stderr, "ERROR: bad format, line\n\t'%s'\n", line);
        return 1;
    }
    if (t < 0 || t >= MAXTRACKS ||
        s < 0 || s >= SECTORS ||
        (tag[0] == 'H' && (b != 2)) ||    // headers are always 2 bytes
        (tag[0] == 'D' && (b != 256)) ||  // data body is always 256 bytes
        e < 0 || e > 1) {
        fprintf(stderr, "ERROR: bad file format, block\n\t'%s'\n", line);
        return 1;
    }

    // now suck down 'b' bytes of ascii hex bytes
    for(i=0; i<b; i++) {
        int byte;
        if (fscanf(fp, "%s", p) != 1) {
            fprintf(stderr, "Error: bad format file\n");
            return 1;
        }
        byte = (int)strtol(p, (char**)NULL, 16);
        if (byte < 0 || byte > 255) {
            fprintf(stderr, "\nError: illegal byte value %s\n", p);
            return 1;
        }
        mem[i] = byte;
    }

    if (tag[0] == 'H') {
        // HEADER
        disk->trk[t].sector[s].hdr_error = e;
        disk->trk[t].sector[s].hdr_bytes = b;
        memcpy( disk->trk[t].sector[s].header, mem, b);
    } else {
        // DATA
        disk->trk[t].sector[s].data_error = e;
        disk->trk[t].sector[s].data_bytes = b;
        memcpy( disk->trk[t].sector[s].data, mem, b);
    }

    // tracks are numbered from 0, but opts.tracks is 1-based
    if (t+1 > opts.tracks)
        opts.tracks = t+1;

    return 0;
}

int
read_byt_file(const char *filename, svh_disk_t *disk)
{
    char line[300];             // holds line to parse
    int track, i;
    int lines = 0;
    FILE *fp;

    if (!file_is_byt(filename))
        return 1;

    opts.tracks = 0;

    fp = fopen(filename, "r");
    assert(fp != NULL);

    while (fgets(line, sizeof(line)-1, fp) != NULL) {

        int idx_start[SECTORS+1];   // starting offset of given track
        int fast, slow;             // mean value of fast and slow edges, from histogram
        int failure;

        lines++;

        strip_line_ending(line);

        // the first line has the magic string
        if ((lines == 1) && (strcmp(line,"HELIOS bytes") == 0))
            continue;

        if (line[0] == '\0' ||  // blank line
            line[0] == '#')     // comment line
            continue;

        if (strncmp(line, "LABEL ", 6) == 0) {
            label_append( &(disk->label), &line[6] );
            continue;
        }

        if ( (strncmp(line, "HEADER ", 7) == 0) ||
             (strncmp(line, "DATA ",   5) == 0) ) {
            if (read_byt_block(fp, line, disk) != 0) {
                fclose(fp);
                return 1;
            }
            continue;
        }

        fprintf(stderr, "ERROR: file format, unknown line\n\t'%s'\n", line);
        fclose(fp);
        return 1;

    } // while()

    return 0;  // OK
}

// ========================================================================
// dump .byt disk image
// return 0 on success, !=0 on failure
// ========================================================================

int
write_byt_file(const char *filename, svh_disk_t *disk)
{
    FILE *fp;
    int t, s, i;

    assert(disk != NULL);

    fp = fopen(filename, "w");
    if (!fp) {
        fprintf(stderr, "Error opening file '%s' for writing\n", filename);
        return 1;
    }

    fprintf(fp, "HELIOS bytes\n");
    label_emit(fp, disk->label);

    for(t=0; t<opts.tracks; t++) {
        for(s=0; s<SECTORS; s++) {

            int hdr_err    = disk->trk[t].sector[s].hdr_error;
            int hdr_bytes  = disk->trk[t].sector[s].hdr_bytes;
            int data_err   = disk->trk[t].sector[s].data_error;
            int data_bytes = disk->trk[t].sector[s].data_bytes;

            // dump header block
            if (hdr_bytes != 0) {
                fprintf(fp, "HEADER %d %d %d %d\n", t, s, hdr_bytes, hdr_err);
                for(i=0; i<hdr_bytes; i++) {
                    fprintf(fp, "%02X ", disk->trk[t].sector[s].header[i]);
                    if (i%16 == 15)
                        fprintf(fp, "\n");
                }
                if (hdr_bytes%16 != 0)
                    fprintf(fp, "\n");
            }

            // dump data block
            if (data_bytes != 0) {
                fprintf(fp, "DATA %d %d %d %d\n", t, s, data_bytes, data_err);
                for(i=0; i<data_bytes; i++) {
                    fprintf(fp, "%02X ", disk->trk[t].sector[s].data[i]);
                    if (i%16 == 15)
                        fprintf(fp, "\n");
                }
                if (data_bytes%16 != 0)
                    fprintf(fp, "\n");
            }

        } // for(s)
    } // for(t)

    fclose(fp);
    return 0;   // success
}


// ========================================================================
// dump .wvd disk image
// return 0 on success, !=0 on failure
// ========================================================================

int
write_wvd_file(const char *filename, svh_disk_t *disk)
{
    FILE *fp;
    int t, s, i;
    int num_sectors = opts.tracks*SECTORS;
    int interleave[16] =
        { 0, 4, 8, 12, 1, 5, 9, 13, 2, 6, 10, 14, 3, 7, 11, 15 };

    assert(disk != NULL);

    fp = fopen(filename, "wb");
    if (!fp) {
        fprintf(stderr, "Error opening file '%s' for writing\n", filename);
        return 1;
    }

    fprintf(fp, "WANG");        // 0-4: magic string
    fputc('\0', fp);
    fprintf(fp, "%c", 0x00);    //   5: write format
    fprintf(fp, "%c", 0x00);    //   6: read  format
    fprintf(fp, "%c", 0x00);    //   7: write protect
    fprintf(fp, "%c%c",         // 8-9: num sectors
                num_sectors % 256,
                num_sectors / 256 );
    fprintf(fp, "%c", 0x01);    //  10: media type: 8" floppy
    for(i=11; i<16; i++)        // 11-15: padding
        fputc('\0', fp);
    i=16;
    while ((disk->label != NULL) && (i < 255) && disk->label[i-16]) {
        fputc(disk->label[i-16], fp);
        i++;
    }
    while (i < 256) {
        fputc('\0', fp);  // pad with 00 / terminate with 00
        i++;
    }

    for(t=0; t<opts.tracks; t++) {
        for(s=0; s<SECTORS; s++) {

            // dump data block
            int from = interleave[s];
            i = fwrite(disk->trk[t].sector[from].data, 1, 256, fp);
            if (i != 256) {
                fprintf(stderr,"Warning: Wrote %d, expected %d bytes "
                               "for track %d, sector %d\n", i, 256, t, s);
            }
        } // for(s)
    } // for(t)

    fclose(fp);
    return 0;   // success
}
