/*
 *    CWWVD -- utility program to read Wang 2200 formatted disks using
 *             a Catweasel card to create virtual disk images
 *
 *    Author: Jim Battle, jim@thebattles.net (as of 12/2011)
 *
 *    Some of this code is based on:
 *        cw2dmk: Dump floppy disk from Catweasel to .dmk format.
 *        Copyright (C) 2000 Timothy Mann
 *        Id: cw2dmk.c,v 1.28 2003/06/19 05:53:28 mann Exp $
 *        Depends on Linux Catweasel driver code by Michael Krause
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *    USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>     // usage() varargs

#include "cwwvd.h"
#include "fileops.h"

#include "cwfloppy.h"
#include "cwpci.h"

static char version_string[] = "10/2011";

// ========================================================================
// utility functions
// ========================================================================

// just like malloc, but exits with error msg if malloc fails
void *
safe_malloc(size_t bytes)
{
    void *p = NULL;

    assert(bytes >= 0);

    p = (void*)malloc(bytes);
    if (p == NULL) {
        fprintf(stderr, "Trouble allocating %d bytes of memory ... sorry!\n", bytes);
        exit(-1);
    }
    return p;
}

// remove trailing spaces, tabs, carriage returns, and newlines
void
strip_line_ending(char *p)
{
    int len;

    assert(p != NULL);
    len = strlen(p);
    while (len > 0 &&
                (p[len-1] == '\n' ||    // newline
                 p[len-1] == '\r' ||    // return
                 p[len-1] == '\t' ||    // tab
                 p[len-1] == ' ') )     // space
        p[--len] = '\0';
}

// return NULL on failure, pointer to static buffer otherwise
// maximum length capture is 200 chars.
char *
get_first_line_of_file(const char *filename)
{
    static char line[200];
    int empty;

    FILE *fp = fopen(filename, "r");
    if (fp == NULL)
        return NULL;

    empty = (fgets(line, sizeof(line)-1, fp) == NULL);
    fclose(fp);
    if (empty)
        return NULL;

    strip_line_ending(line);
    return line;
}


// append the source string to the destination string.
// it is assumed that the appended string does not have \n,
// and that the original string doesn't end in \n, so we
// put one between them.
void
label_append(char **d, char *s)
{
    char *p = NULL;
    int dlen, slen;

    assert(s != NULL);
    assert(d != NULL);

    // allocate space to hold both strings
    slen = strlen(s);
    dlen = (*d == NULL) ? 0 : strlen(*d);
    p = safe_malloc(slen + dlen + 3);  // +3 for \r\n\0

    if (*d == NULL) {
        // first line of label
        strcpy(p, s);
    } else {
        // concatenate them
        strcpy(p, *d);
        strcpy(p+dlen, "\r\n");
        strcpy(p+dlen+1, s);
    }

    // free old one, replace with new one
    if (*d == NULL) {
        *d = p;
    } else {
        free(*d);
        *d = p;
    }
}

// send "LABEL blah blah blah" lines to specified RAW or BYT file
void
label_emit(FILE *fp, char *label)
{
    char *s, *e, *p;    // start, end
    int len;

    assert(fp != NULL);

    if (label == NULL)  // no label on disk
        return;
    len = strlen(label);

    s = &label[0];
    e = &label[len];    // points at terminating \0 byte
    for(p=s; p<e; p++) {
        if ( (*p == '\r') ||    // end of a line
             (*p == '\n') ||    // end of a line
             (p == e-1)) {      // in case it doesn't end in \n
            int k = p-s+1         // length of label line ...
                  - (*p == '\r'); // ... not counting \r
                  - (*p == '\n'); // ... not counting \n
            char *m = safe_malloc(k+1); // +1 for null
            if (k > 0)
                strncpy(m, s, k);
            m[k] = '\0';
            fprintf(fp, "LABEL %s\n", m);
            free(m);
            if (p[0] == '\r' && p[1] == '\n') // skip pair
                p++;
            s = p+1;    // mark start of next line
        }
    }
}


// ========================================================================
// svh_disk_t operations
// ========================================================================

void
disk_init(svh_disk_t *disk)
{
    int t, s;

    disk->label = NULL;

    for(t=0; t<opts.tracks; t++) {
        trk_t *trk = &(disk->trk[t]);
        for(s=0; s<SECTORS; s++) {
            trk->sector[s].samples = 0;
            trk->sector[s].raw = NULL;
            trk->sector[s].hdr_bytes = 0;
            trk->sector[s].hdr_error = 0;
            trk->sector[s].data_bytes = 0;
            trk->sector[s].data_error = 0;
        }
    }
}

void
disk_free(svh_disk_t *disk)
{
    int t, s;

    if (disk->label != NULL)
        free(disk->label);

    for(t=0; t<opts.tracks; t++)
    for(s=0; s<SECTORS; s++) {
        if (disk->trk[t].sector[s].raw != NULL)
            free(disk->trk[t].sector[s].raw);
    }

    disk_init(disk);
}

// ========================================================================
// Debug getter
// ========================================================================

static int debug = 0;

int
get_debug() {
    return(debug);
}

// ========================================================================
// Debug setter
// ========================================================================

void
set_debug(int value) {
    debug = value;
}

// ========================================================================
// usage help
// ========================================================================

void
usage(const char *fmt, ...)
{
    char buff[1000];
    va_list args;

    // print optional reason for error message
    if (fmt != NULL) {
        va_start(args, fmt);
        vsprintf(buff, fmt, args);
        va_end(args);
        fprintf(stderr, "Error: %s\n", buff);
    }

    fprintf(stderr,
        "command summary (arguments after <src> can come in any order):\n"
        "    cwwvd -version   -- report program version\n"
        "    cwwvd <src> <opts> [-r foo.raw] [-b foo.byt] [-w foo.wvd]\n"
        "    <src> is one of:\n"
        "        /n           -- disk drive n (0-3) on the attached catweasel card\n"
        "        filename.raw -- unprocessed dump of bit transition timings, by track\n"
        "        filename.byt -- header and data bytes FM decoded by track and sector\n"
        "    <opts> is zero or more of:\n"
        "        -h           -- print this help information\n"
        "        -d<n>        -- specify the debug logging level (default 0; 4=noisy)\n"
        "        -l \"label\"   -- add a line to the label metadata\n"
        "        -retries <#> -- number of attempts to read a track (10 default)\n"
        "        -tracks <#>  -- number of tracks to decode (77 default)\n"
        "Notes:\n"
        "    A .raw file can only be produced from a real disk.\n"
        "    Use multiple \"-l\"s for multiline labels.\n"
    );
    exit(-1);
}

// ========================================================================
// main routine
// ========================================================================

// command line options
// this is a global so decoding routines can access tuning parameters
// without having to pass them around
opts_t opts;

int
main(int argc, char** argv)
{
    char *pSrcFile   = NULL;    // source file to read (NULL if disk)
    char *pRFile     = NULL;    // name to save .RAW file to
    char *pBFile     = NULL;    // name to save .BYT file to
    char *pWFile     = NULL;    // name to save .WVD file to
    char *pLabel     = NULL;    // label specified on command line
    int   disk_unit  = -1;      // >=0 indicates to read from disk
    int   rv = 0;               // return value from program
    int   i;
    svh_disk_t *disk = NULL;    // holds decoded disk

    // set default option values
    sector_decode_defaults();
    opts.retries = 10;
    opts.tracks = 77;   // some disks have only 64 meaningful tracks

    // ====== parse arguments ======

    // don't insist on having a source if all they want is the version
    if (argc > 1 && (!strcmp(argv[1], "-v") || !strcmp(argv[1], "-version"))) {
        fprintf(stderr, "cwwvd version %s\n", version_string);
        exit(0);
    }

    // we need to have at least these args:
    //   program name (argv[0]), source, dest
    if (argc < 3) {
        usage(NULL);
    }

    // see if a drive (/0 /1 /2 /3) was specified
    i = 1;
    if ((argv[i][0] == '/') && (strlen(argv[i]) == 2)) {
        if (argv[i][1] < '0' || argv[i][1] > '3')
            usage("Disk unit must be from 0 to 3, not %c", argv[i][1]);
        disk_unit = argv[i][1] - '0';
        i++;
    }

    // if "disk" wasn't specified, it must be a file name
    if (disk_unit < 0) {
        pSrcFile = argv[i];
        FILE *fp = fopen(pSrcFile, "r");
        if (fp == NULL) {
            usage("Couldn't open file '%s' for reading.", pSrcFile);
        }
        i++;
    }

    // scan for -flags
    while (i < argc) {

        // specify raw output file
        if (!strcmp(argv[i], "-r")) {
            if (pRFile != NULL)
                usage("-r flag can appear only once");
            if (++i >= argc)
                usage("-r must be followed by a target file name");
            pRFile = argv[i];   // grab associated filename
            i++;
            continue;
        }

        // specify decoded byte output file
        if (!strcmp(argv[i], "-b")) {
            if (pBFile != NULL)
                usage("-b flag can appear only once");
            if (++i >= argc)
                usage("-b must be followed by a target file name");
            pBFile = argv[i];   // grab associated filename
            i++;
            continue;
        }

        // specify wang virtual disk output file
        if (!strcmp(argv[i], "-w")) {
            if (++i >= argc) {
                usage("-i must be followed by a target file name");
            }
            pWFile = argv[i];
            ++i;
            continue;
        }

        // help (usage) message
        if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "-help")) {
            usage(NULL);
        }

        if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "-version")) {
            fprintf(stderr, "cwwvd version %s\n", version_string);
            exit(0);
        }

        // debug; allow either -d<n> or -d <n>
        if (!strncmp(argv[i], "-d", 2)) {
            if (strlen(argv[i]) == 2) {
                set_debug(get_debug() + 1);
            } else if (argv[i][2] >= '0' && argv[i][2] <= '9') {
                set_debug((int)strtol(&argv[i][2], NULL, 10));
            } else {
                usage("-d or -d<digits>");
            }
            ++i;
            continue;
        }

        // number of tracks to capture
        if (!strcmp(argv[i], "-retries")) {
            if (++i >= argc)
                usage("-retries must be followed by an integer (eg, 10)");
            opts.retries = atoi(argv[i]);
            i++;
            continue;
        }

        // number of tracks to capture
        if (!strcmp(argv[i], "-tracks")) {
            if (++i >= argc)
                usage("-tracks must be followed by an integer (eg, 64)");
            opts.tracks = atoi(argv[i]);
            if (opts.tracks > MAXTRACKS)
                usage("-tracks must be less than %d", MAXTRACKS);
            i++;
            continue;
        }

        // set the number of symbols to skip after index mark before
        // starting header preamble sync
        if (!strcmp(argv[i], "-skip1")) {
            if (++i >= argc)
                usage("-skip1 must be followed by an integer (eg, 76)");
            opts.skip1 = atoi(argv[i]);
            i++;
            continue;
        }

        // set the number of symbols to skip after header before
        // starting data preamble sync
        if (!strcmp(argv[i], "-skip2")) {
            if (++i >= argc)
                usage("-skip2 must be followed by an integer (eg, 140)");
            opts.skip2 = atoi(argv[i]);
            i++;
            continue;
        }

        // specify line of label
        if (!strcmp(argv[i], "-l")) {
            if (++i >= argc)
                usage("-l must be followed by a label string");
            label_append(&pLabel, argv[i]);
            i++;
            continue;
        }

        usage("Unknown flag '%s'", argv[i]);
    }

    if ((pRFile != NULL) && (disk_unit < 0))
        usage("Raw files can only be extracted from real disks");

    // check the argument parsing
    if (get_debug()) {
        if (disk_unit >= 0)
        fprintf(stderr, "disk_unit=%d\n", disk_unit);
        fprintf(stderr, "debug: %d\n",get_debug());
        if (pSrcFile != NULL) 
            fprintf(stderr, "source file is '%s'\n", pSrcFile);
        if (pRFile != NULL) fprintf(stderr, "-r file is '%s'\n", pRFile);
        if (pBFile != NULL) fprintf(stderr, "-b file is '%s'\n", pBFile);
        if (pWFile != NULL) fprintf(stderr, "-w file is '%s'\n", pWFile);
        fprintf(stderr, "-tracks=%d, -retries=%d\n", opts.tracks, opts.retries);
        fprintf(stderr, "-skip1=%d, -skip2=%d\n", opts.skip1, opts.skip2);
        label_emit(stderr, pLabel);
    }

    // ====== carry out requested action ======

    disk = (svh_disk_t*)safe_malloc(sizeof(svh_disk_t));

    // start with an empty disk data structure
    disk_init(disk);

    // get the starting data
    if (pSrcFile == NULL) {
        // no source file supplied, so a real disk is assumed
        rv = capture_disk(disk_unit, disk);
    } else {
        if (file_is_raw(pSrcFile)) {
            rv = read_raw_file(pSrcFile, disk);
        } else if (file_is_byt(pSrcFile)) {
            rv = read_byt_file(pSrcFile, disk);
        } else {
            fprintf(stderr, "ERROR: can't determine file format of '%s'\n", pSrcFile);
            exit(-1);
        }
    }
    if (rv != 0) {
        // exit(rv);
        printf("WARNING: disk image might not be entirely valid\n");
    }

    // at this point we have filled the "disk" structure with the disk image

    // if the user specified a label for the disk, use that instead
    if (pLabel != NULL) {
        if (disk->label != NULL)
            free(disk->label);
        disk->label = pLabel;
    }

    if (pRFile != NULL) {
        // dump raw disk image
        rv = write_raw_file(pRFile, disk);
        if (rv != 0) {
            fprintf(stderr, "Error dumping .raw file -- aborting!\n");
            exit(rv);
        }
    }

    if (pBFile != NULL) {
        // dump to decoded byte file
        rv = write_byt_file(pBFile, disk);
        if (rv != 0) {
            fprintf(stderr, "Error dumping .byt file -- aborting!\n");
            exit(rv);
        }
    }

    if (pWFile != NULL) {
        // dump to decoded wang virtual disk image file
        rv = write_wvd_file(pWFile, disk);
        if (rv != 0) {
            fprintf(stderr, "Error dumping .wvd file -- aborting!\n");
            exit(rv);
        }
    }

    disk_free(disk);
    disk = NULL;

    return 0;
}
